'use strict';

var _objectSpread = require('@babel/runtime/helpers/objectSpread2');
var _defineProperty = require('@babel/runtime/helpers/defineProperty');
var baseControllers = require('@toruslabs/base-controllers');
var constants$1 = require('@toruslabs/constants');
var fndBase = require('@toruslabs/fnd-base');
var sessionManager = require('@toruslabs/session-manager');
var torus_js = require('@toruslabs/torus.js');
var auth = require('@web3auth/auth');
var base = require('@web3auth/base');
var bs58 = require('bs58');
var asyncStorage = require('./asyncStorage.js');
var constants = require('./constants.js');
var utils = require('./utils.js');

class Web3Auth extends auth.SafeEventEmitter {
  constructor(options) {
    var _options$chainConfig, _options$privateKeyPr, _options$chainConfig2, _options$useDkg;
    super();
    _defineProperty(this, "coreOptions", void 0);
    _defineProperty(this, "connectedAdapterName", base.WALLET_ADAPTERS.SFA);
    _defineProperty(this, "SFA_ISSUER", "https://authjs.web3auth.io/jwks");
    _defineProperty(this, "status", base.ADAPTER_STATUS.NOT_READY);
    _defineProperty(this, "authInstance", null);
    _defineProperty(this, "state", {});
    _defineProperty(this, "nodeDetails", null);
    _defineProperty(this, "torusPrivKey", null);
    _defineProperty(this, "privKeyProvider", null);
    _defineProperty(this, "accountAbstractionProvider", null);
    _defineProperty(this, "sessionManager", void 0);
    _defineProperty(this, "currentStorage", void 0);
    _defineProperty(this, "baseStorageKey", "sfa_store");
    _defineProperty(this, "sessionNamespace", "sfa");
    _defineProperty(this, "plugins", {});
    if (!options.clientId) throw base.WalletInitializationError.invalidParams("Please provide a valid clientId in constructor");
    if (!options.privateKeyProvider) throw base.WalletInitializationError.invalidParams("Please provide a valid privateKeyProvider in constructor");
    if ((_options$chainConfig = options.chainConfig) !== null && _options$chainConfig !== void 0 && _options$chainConfig.chainId && ((_options$privateKeyPr = options.privateKeyProvider.currentChainConfig) === null || _options$privateKeyPr === void 0 ? void 0 : _options$privateKeyPr.chainId) !== ((_options$chainConfig2 = options.chainConfig) === null || _options$chainConfig2 === void 0 ? void 0 : _options$chainConfig2.chainId)) {
      throw base.WalletInitializationError.invalidParams("chainId in privateKeyProvider and chainConfig should be same");
    }
    this.coreOptions = _objectSpread(_objectSpread({}, options), {}, {
      web3AuthNetwork: options.web3AuthNetwork || auth.WEB3AUTH_NETWORK.MAINNET,
      sessionTime: options.sessionTime || 86400,
      storageServerUrl: options.storageServerUrl || "https://session.web3auth.io",
      storageKey: options.storageKey || "local",
      chainConfig: options.chainConfig || options.privateKeyProvider.currentChainConfig,
      storage: this.getStorage(options.storage),
      mode: options.mode || constants.SDK_MODE.WEB
    });
    this.coreOptions.useDkg = this.isSapphireNetwork() ? (_options$useDkg = options.useDkg) !== null && _options$useDkg !== void 0 ? _options$useDkg : true : true;
    this.privKeyProvider = options.privateKeyProvider;
    if (options.accountAbstractionProvider) {
      this.accountAbstractionProvider = options.accountAbstractionProvider;
    }
  }
  get connected() {
    return this.status === base.ADAPTER_STATUS.CONNECTED;
  }
  get provider() {
    if (this.status !== base.ADAPTER_STATUS.NOT_READY && this.privKeyProvider) {
      var _this$accountAbstract;
      return (_this$accountAbstract = this.accountAbstractionProvider) !== null && _this$accountAbstract !== void 0 ? _this$accountAbstract : this.privKeyProvider;
    }
    return null;
  }
  set provider(_) {
    throw new Error("Not implemented");
  }
  async init() {
    if (this.status !== base.ADAPTER_STATUS.NOT_READY) throw base.WalletInitializationError.notReady("Already initialized");
    if (!this.privKeyProvider.currentChainConfig || !this.privKeyProvider.currentChainConfig.chainNamespace || !this.privKeyProvider.currentChainConfig.chainId) {
      throw base.WalletInitializationError.invalidParams("provider should have chainConfig and should be initialized with chainId and chainNamespace");
    }
    const storageKey = this.coreOptions.sessionKey || `${this.baseStorageKey}_${this.coreOptions.usePnPKey ? "pnp" : "core_kit"}`;
    this.currentStorage = new asyncStorage.AsyncStorage(storageKey, this.coreOptions.storage);
    this.nodeDetails = fndBase.fetchLocalConfig(this.coreOptions.web3AuthNetwork, constants$1.KEY_TYPE.SECP256K1);
    this.authInstance = new torus_js.Torus({
      clientId: this.coreOptions.clientId,
      enableOneKey: true,
      network: this.coreOptions.web3AuthNetwork,
      serverTimeOffset: this.coreOptions.serverTimeOffset
    });
    const sessionId = await this.currentStorage.get("sessionId");
    this.sessionManager = new sessionManager.SessionManager({
      sessionServerBaseUrl: this.coreOptions.storageServerUrl,
      sessionTime: this.coreOptions.sessionTime,
      sessionNamespace: this.sessionNamespace,
      sessionId
    });
    this.subscribeToEvents();
    if (this.plugins[constants.PASSKEYS_PLUGIN]) {
      await this.plugins[constants.PASSKEYS_PLUGIN].initWithWeb3Auth(this);
    }

    // if sessionId exists in storage, then try to rehydrate session.
    if (sessionId) {
      // we are doing this to make sure sessionKey is set
      // before we call authorizeSession in both cjs and esm bundles.
      this.sessionManager.sessionId = sessionId;
      const data = await this.sessionManager.authorizeSession().catch(() => {
        this.currentStorage.set("sessionId", "");
      });
      if (data && data.privKey) {
        this.torusPrivKey = data.basePrivKey;
        const finalPrivKey = await this.getFinalPrivKey(data.basePrivKey);
        await this.privKeyProvider.setupProvider(finalPrivKey);
        // setup aa provider after private key provider is setup
        if (this.accountAbstractionProvider) {
          await this.accountAbstractionProvider.setupProvider(this.privKeyProvider);
        }
        this.updateState(data);
        this.status = base.ADAPTER_STATUS.CONNECTED;
        this.emit(base.ADAPTER_EVENTS.CONNECTED, {
          adapter: this.connectedAdapterName,
          provider: this.provider,
          reconnected: true
        });
      }
    }
    if (!this.state.privKey) {
      this.status = base.ADAPTER_STATUS.READY;
      this.emit(base.ADAPTER_EVENTS.READY, this.connectedAdapterName);
    }
  }
  async authenticateUser() {
    if (!this.connected) throw base.WalletLoginError.notConnectedError();
    const {
      userInfo
    } = this.state;
    if (userInfo !== null && userInfo !== void 0 && userInfo.idToken) return {
      idToken: userInfo.idToken
    };
    const {
      chainNamespace,
      chainId
    } = this.coreOptions.chainConfig || {};
    if (!this.authInstance || !this.privKeyProvider) throw base.WalletInitializationError.notReady();
    const [accounts, publicKey] = await Promise.all([this.privKeyProvider.provider.request({
      method: chainNamespace === base.CHAIN_NAMESPACES.EIP155 ? "eth_accounts" : "getAccounts"
    }), this.privKeyProvider.provider.request({
      method: "public_key"
    })]);
    // const thresholdPrivKey = this._getBasePrivKey();

    if (accounts && accounts.length > 0) {
      const existingToken = base.getSavedToken(accounts[0], "SFA");
      if (existingToken) {
        const isExpired = base.checkIfTokenIsExpired(existingToken);
        if (!isExpired) {
          return {
            idToken: existingToken
          };
        }
      }
      const payload = {
        domain: typeof window !== "undefined" ? window.location.origin : "reactnative",
        uri: typeof window !== "undefined" ? window.location.href : "com://reactnative",
        address: accounts[0],
        chainId: parseInt(chainId, 16),
        version: "1",
        nonce: Math.random().toString(36).slice(2),
        issuedAt: new Date().toISOString()
      };
      const additionalMetadata = {
        email: userInfo.email,
        name: userInfo.name,
        profileImage: userInfo.profileImage,
        aggregateVerifier: userInfo.aggregateVerifier,
        verifier: userInfo.verifier,
        verifierId: userInfo.verifierId,
        typeOfLogin: userInfo.typeOfLogin || "jwt",
        oAuthIdToken: userInfo.oAuthIdToken,
        oAuthAccessToken: userInfo.oAuthAccessToken,
        wallets: [],
        signatures: this.state.signatures,
        network: this.coreOptions.web3AuthNetwork
      };
      if (this.coreOptions.usePnPKey) {
        additionalMetadata.wallets.push({
          public_key: publicKey,
          type: "web3auth_app_key",
          curve: chainNamespace === base.CHAIN_NAMESPACES.EIP155 ? constants$1.KEY_TYPE.SECP256K1 : constants$1.KEY_TYPE.ED25519
        });
      } else {
        additionalMetadata.wallets.push({
          public_key: publicKey,
          type: "web3auth_threshold_key",
          curve: chainNamespace === base.CHAIN_NAMESPACES.EIP155 ? constants$1.KEY_TYPE.SECP256K1 : constants$1.KEY_TYPE.ED25519
        });
      }
      const challenge = await baseControllers.signChallenge(payload, chainNamespace);
      const signedMessage = await this._getSignedMessage(challenge, accounts, chainNamespace);
      const idToken = await baseControllers.verifySignedChallenge(chainNamespace, signedMessage, challenge, this.SFA_ISSUER, this.coreOptions.sessionTime, this.coreOptions.clientId, this.coreOptions.web3AuthNetwork, undefined, additionalMetadata);
      base.saveToken(accounts[0], "SFA", idToken);
      return {
        idToken
      };
    }
  }
  async addChain(chainConfig) {
    if (this.status === base.ADAPTER_STATUS.NOT_READY) throw base.WalletInitializationError.notReady("Please call init first.");
    return this.privKeyProvider.addChain(chainConfig);
  }
  switchChain(params) {
    if (this.status === base.ADAPTER_STATUS.NOT_READY) throw base.WalletInitializationError.notReady("Please call init first.");
    return this.privKeyProvider.switchChain(params);
  }
  async getPostboxKey(loginParams) {
    if (this.status === base.ADAPTER_STATUS.NOT_READY) throw base.WalletInitializationError.notReady("Please call init first.");
    return this.getTorusKey(loginParams);
  }

  /**
   * Use this function only with verifiers created on developer dashboard (https://dashboard.web3auth.io)
   * @param loginParams - Params used to login
   * @returns provider to connect
   */
  async connect(loginParams) {
    if (this.status === base.ADAPTER_STATUS.CONNECTED) throw base.WalletLoginError.connectionError("Already connected. Please check status before calling connect.");
    if (this.status !== base.ADAPTER_STATUS.READY) throw base.WalletInitializationError.notReady("Please call init first.");
    const {
      verifier,
      verifierId,
      idToken,
      subVerifierInfoArray
    } = loginParams;
    if (!verifier || !verifierId || !idToken) throw base.WalletInitializationError.invalidParams("verifier or verifierId or idToken  required");
    const {
      torusNodeEndpoints,
      torusIndexes,
      torusNodePub
    } = this.nodeDetails;
    if (loginParams.serverTimeOffset) {
      this.authInstance.serverTimeOffset = loginParams.serverTimeOffset;
    }
    let finalIdToken = idToken;
    let finalVerifierParams = {
      verifier_id: verifierId
    };
    if (subVerifierInfoArray && (subVerifierInfoArray === null || subVerifierInfoArray === void 0 ? void 0 : subVerifierInfoArray.length) > 0) {
      const aggregateVerifierParams = {
        verify_params: [],
        sub_verifier_ids: [],
        verifier_id: ""
      };
      const aggregateIdTokenSeeds = [];
      for (let index = 0; index < subVerifierInfoArray.length; index += 1) {
        const userInfo = subVerifierInfoArray[index];
        aggregateVerifierParams.verify_params.push({
          verifier_id: verifierId,
          idtoken: userInfo.idToken
        });
        aggregateVerifierParams.sub_verifier_ids.push(userInfo.verifier);
        aggregateIdTokenSeeds.push(userInfo.idToken);
      }
      aggregateIdTokenSeeds.sort();
      finalIdToken = torus_js.keccak256(Buffer.from(aggregateIdTokenSeeds.join(String.fromCharCode(29)), "utf8")).slice(2);
      aggregateVerifierParams.verifier_id = verifierId;
      finalVerifierParams = aggregateVerifierParams;
    }
    const retrieveSharesResponse = await this.authInstance.retrieveShares({
      endpoints: torusNodeEndpoints,
      indexes: torusIndexes,
      verifier,
      verifierParams: finalVerifierParams,
      idToken: finalIdToken,
      nodePubkeys: torusNodePub,
      checkCommitment: false,
      useDkg: this.coreOptions.useDkg,
      extraParams: {}
    });
    if (retrieveSharesResponse.metadata.upgraded) {
      throw base.WalletLoginError.mfaEnabled();
    }
    const {
      finalKeyData,
      oAuthKeyData
    } = retrieveSharesResponse;
    const privKey = finalKeyData.privKey || oAuthKeyData.privKey;
    this.torusPrivKey = privKey;
    if (!privKey) throw base.WalletLoginError.fromCode(5000, "Unable to get private key from torus nodes");

    // we are using the original private key so that we can retrieve other keys later on
    let decodedUserInfo;
    try {
      decodedUserInfo = utils.decodeToken(idToken).payload;
    } catch (error) {
      decodedUserInfo = loginParams.fallbackUserInfo;
    }
    const userInfo = {
      name: decodedUserInfo.name || decodedUserInfo.nickname || "",
      email: decodedUserInfo.email || "",
      profileImage: decodedUserInfo.picture || "",
      verifierId,
      verifier,
      typeOfLogin: "jwt",
      oAuthIdToken: idToken
    };
    const signatures = this.getSessionSignatures(retrieveSharesResponse.sessionData);
    await this._finalizeLogin({
      privKey,
      userInfo,
      signatures
    });
    return this.provider;
  }
  async logout() {
    if (!this.connected) throw base.WalletLoginError.notConnectedError("Not logged in");
    if (this.coreOptions.mode !== constants.SDK_MODE.NODE) {
      await this.sessionManager.invalidateSession();
      this.currentStorage.set("sessionId", "");
    }
    this.updateState({
      privKey: "",
      userInfo: {
        name: "",
        email: "",
        profileImage: "",
        verifierId: "",
        verifier: "",
        typeOfLogin: ""
      }
    });
    this.status = base.ADAPTER_STATUS.READY;
    this.emit(base.ADAPTER_EVENTS.DISCONNECTED);
  }
  async getUserInfo() {
    if (!this.connected) throw base.WalletLoginError.userNotLoggedIn();
    return this.state.userInfo;
  }
  addPlugin(plugin) {
    if (this.isRNOrNodeMode()) throw base.WalletInitializationError.invalidParams("Plugins are not supported in this mode");
    if (this.plugins[plugin.name]) throw base.WalletInitializationError.duplicateAdapterError(`Plugin ${plugin.name} already exist`);
    this.plugins[plugin.name] = plugin;
    if (this.status === base.ADAPTER_STATUS.CONNECTED && this.connectedAdapterName) {
      // web3auth is already connected. can initialize plugins
      this.connectToPlugins();
    }
    if (plugin.name === constants.PASSKEYS_PLUGIN && (this.status === base.ADAPTER_STATUS.READY || this.status === base.ADAPTER_STATUS.CONNECTED)) {
      // this does return a Promise but we don't need to wait here.
      // as its mostly a sync function.
      this.plugins[constants.PASSKEYS_PLUGIN].initWithWeb3Auth(this);
    }
    return this;
  }
  getPlugin(name) {
    if (this.isRNOrNodeMode()) throw base.WalletInitializationError.invalidParams("Plugins are not supported in this mode");
    return this.plugins[name] || null;
  }
  async _finalizeLogin(params) {
    const {
      privKey,
      signatures = [],
      passkeyToken = ""
    } = params;
    this.torusPrivKey = privKey;
    // update the provider with the private key.
    const finalPrivKey = await this.getFinalPrivKey(privKey);
    await this.privKeyProvider.setupProvider(finalPrivKey);
    // setup aa provider after private key provider is setup
    if (this.accountAbstractionProvider) {
      await this.accountAbstractionProvider.setupProvider(this.privKeyProvider);
    }

    // We only need to authenticate user in web mode.
    // This is for the passkey plugin mode only.
    if (this.coreOptions.mode === constants.SDK_MODE.WEB) {
      const {
        idToken
      } = await this.authenticateUser().catch(_ => ({
        idToken: ""
      }));
      if (params.userInfo) {
        params.userInfo.idToken = idToken;
      } else {
        params.userInfo = {
          idToken
        };
      }
    }

    // We dont need to save the session in node mode.
    if (this.coreOptions.mode !== constants.SDK_MODE.NODE) {
      // save the data in the session.
      const sessionId = sessionManager.SessionManager.generateRandomSessionKey();
      this.sessionManager.sessionId = sessionId;
      await this.sessionManager.createSession({
        basePrivKey: privKey,
        privKey: finalPrivKey,
        userInfo: params.userInfo,
        signatures,
        passkeyToken
      });
      this.currentStorage.set("sessionId", sessionId);
    }

    // update the local state.
    this.updateState({
      privKey: finalPrivKey,
      basePrivKey: privKey,
      userInfo: params.userInfo,
      signatures,
      passkeyToken
    });
    this.status = base.ADAPTER_STATUS.CONNECTED;
    this.emit(base.ADAPTER_EVENTS.CONNECTED, {
      adapter: this.connectedAdapterName,
      provider: this.provider,
      reconnected: false
    });
  }
  _getBasePrivKey() {
    return this.torusPrivKey;
  }
  updateState(newState) {
    this.state = _objectSpread(_objectSpread({}, this.state), newState);
  }
  async getFinalPrivKey(privKey) {
    let finalPrivKey = privKey.padStart(64, "0");
    // get app scoped keys.
    if (this.coreOptions.usePnPKey) {
      finalPrivKey = this.getSubKey(finalPrivKey);
    }
    if (this.coreOptions.chainConfig.chainNamespace === base.CHAIN_NAMESPACES.SOLANA) {
      finalPrivKey = this.getEd25519Key(finalPrivKey);
    }
    return finalPrivKey;
  }
  getSubKey(privKey) {
    const pnpPrivKey = auth.subkey(privKey, Buffer.from(this.coreOptions.clientId, "base64"));
    return pnpPrivKey.padStart(64, "0");
  }
  getEd25519Key(privKey) {
    if (!this.privKeyProvider.getEd25519Key) {
      throw base.WalletLoginError.fromCode(5000, "Private key provider is not valid, Missing getEd25519Key function");
    }
    return this.privKeyProvider.getEd25519Key(privKey);
  }
  async getTorusKey(loginParams) {
    const {
      verifier,
      verifierId,
      idToken,
      subVerifierInfoArray
    } = loginParams;
    const {
      torusNodeEndpoints,
      torusIndexes,
      torusNodePub
    } = fndBase.fetchLocalConfig(this.coreOptions.web3AuthNetwork, constants$1.KEY_TYPE.SECP256K1);
    if (loginParams.serverTimeOffset) {
      this.authInstance.serverTimeOffset = loginParams.serverTimeOffset;
    }
    let finalIdToken = idToken;
    let finalVerifierParams = {
      verifier_id: verifierId
    };
    if (subVerifierInfoArray && (subVerifierInfoArray === null || subVerifierInfoArray === void 0 ? void 0 : subVerifierInfoArray.length) > 0) {
      const aggregateVerifierParams = {
        verify_params: [],
        sub_verifier_ids: [],
        verifier_id: ""
      };
      const aggregateIdTokenSeeds = [];
      for (let index = 0; index < subVerifierInfoArray.length; index += 1) {
        const userInfo = subVerifierInfoArray[index];
        aggregateVerifierParams.verify_params.push({
          verifier_id: verifierId,
          idtoken: userInfo.idToken
        });
        aggregateVerifierParams.sub_verifier_ids.push(userInfo.verifier);
        aggregateIdTokenSeeds.push(userInfo.idToken);
      }
      aggregateIdTokenSeeds.sort();
      finalIdToken = torus_js.keccak256(Buffer.from(aggregateIdTokenSeeds.join(String.fromCharCode(29)), "utf8")).slice(2);
      aggregateVerifierParams.verifier_id = verifierId;
      finalVerifierParams = aggregateVerifierParams;
    }
    const retrieveSharesResponse = await this.authInstance.retrieveShares({
      endpoints: torusNodeEndpoints,
      indexes: torusIndexes,
      verifier,
      verifierParams: finalVerifierParams,
      idToken: finalIdToken,
      useDkg: this.coreOptions.useDkg,
      checkCommitment: false,
      extraParams: {},
      nodePubkeys: torusNodePub
    });
    const postboxKey = torus_js.Torus.getPostboxKey(retrieveSharesResponse);
    return postboxKey.padStart(64, "0");
  }
  isSapphireNetwork() {
    return Object.values(constants$1.TORUS_SAPPHIRE_NETWORK).includes(this.coreOptions.web3AuthNetwork);
  }
  getSessionSignatures(sessionData) {
    return sessionData.sessionTokenData.filter(i => Boolean(i)).map(session => JSON.stringify({
      data: session.token,
      sig: session.signature
    }));
  }
  subscribeToEvents() {
    this.on(base.ADAPTER_EVENTS.CONNECTED, async () => {
      this.connectToPlugins();
    });
    this.on(base.ADAPTER_EVENTS.DISCONNECTED, async () => {
      const localPlugins = Object.values(this.plugins).filter(plugin => plugin.name !== constants.PASSKEYS_PLUGIN);
      localPlugins.forEach(async plugin => {
        try {
          if (!plugin.SUPPORTED_ADAPTERS.includes(base.WALLET_ADAPTERS.SFA)) {
            return;
          }
          await plugin.disconnect();
        } catch (error) {
          // swallow error if connector adapter doesn't supports this plugin.
          if (error.code === 5211) {
            return;
          }
          base.log.error(error);
        }
      });
    });
  }
  connectToPlugins() {
    const localPlugins = Object.values(this.plugins).filter(plugin => plugin.name !== constants.PASSKEYS_PLUGIN);
    localPlugins.forEach(async plugin => {
      try {
        if (!plugin.SUPPORTED_ADAPTERS.includes(base.WALLET_ADAPTERS.SFA)) {
          return;
        }
        if (plugin.status === base.PLUGIN_STATUS.CONNECTED) return;
        await plugin.initWithWeb3Auth(this);
        await plugin.connect({
          sessionId: this.sessionManager.sessionId,
          sessionNamespace: this.sessionNamespace
        });
      } catch (error) {
        // swallow error if connector adapter doesn't supports this plugin.
        if (error.code === 5211) {
          return;
        }
        base.log.error(error);
      }
    });
  }
  async _getSignedMessage(challenge, accounts, chainNamespace) {
    const signedMessage = await this.privKeyProvider.provider.request({
      method: chainNamespace === base.CHAIN_NAMESPACES.EIP155 ? "personal_sign" : "signMessage",
      params: chainNamespace === base.CHAIN_NAMESPACES.EIP155 ? [challenge, accounts[0]] : {
        message: Buffer.from(challenge)
      }
    });
    if (chainNamespace === base.CHAIN_NAMESPACES.SOLANA) return bs58.encode(signedMessage);
    return signedMessage;
  }
  isRNOrNodeMode() {
    return this.coreOptions.mode === constants.SDK_MODE.REACT_NATIVE || this.coreOptions.mode === constants.SDK_MODE.NODE;
  }
  getStorage(storage) {
    if (typeof window !== "undefined") {
      if (!storage || storage === "local") return window.localStorage;
      if (storage === "session") return window.sessionStorage;
    } else if (!storage || typeof storage === "string") return new auth.MemoryStore();
    return storage;
  }
}

exports.Web3Auth = Web3Auth;
