'use strict';

var _defineProperty = require('@babel/runtime/helpers/defineProperty');

class AsyncStorage {
  constructor(storeKey, storage) {
    _defineProperty(this, "storage", void 0);
    _defineProperty(this, "_storeKey", void 0);
    this.storage = storage;
    this._storeKey = storeKey;
  }
  async toJSON() {
    if ("getItemAsync" in this.storage) {
      return this.storage.getItemAsync(this._storeKey, {});
    }
    const result = await this.storage.getItem(this._storeKey);
    return result;
  }
  async resetStore() {
    const currStore = await this.getStore();
    await this.setStore({});
    return currStore;
  }
  async getStore() {
    if ("getItemAsync" in this.storage) {
      return JSON.parse((await this.storage.getItemAsync(this._storeKey, {})) || "{}");
    }
    return JSON.parse((await this.storage.getItem(this._storeKey)) || "{}");
  }
  async setStore(store) {
    if ("setItemAsync" in this.storage) {
      return this.storage.setItemAsync(this._storeKey, JSON.stringify(store), {});
    }
    return this.storage.setItem(this._storeKey, JSON.stringify(store));
  }
  async get(key) {
    const store = await this.getStore();
    return store[key];
  }
  async set(key, value) {
    const store = await this.getStore();
    store[key] = value;
    return this.setStore(store);
  }
  async remove(key) {
    const store = await this.getStore();
    delete store[key];
    return this.setStore(store);
  }
}

exports.AsyncStorage = AsyncStorage;
