import { Torus } from "@toruslabs/torus.js";
import { AuthUserInfo, IStorage, SafeEventEmitter } from "@web3auth/auth";
import { CustomChainConfig, IPlugin, IProvider } from "@web3auth/base";
import { ADAPTER_STATUS_TYPE, IAsyncStorage, IFinalizeLoginParams, ISecureStore, IWeb3Auth, LoginParams, SessionData, UserAuthInfo, Web3AuthOptions, Web3AuthSfaEvents } from "./interface";
export declare class Web3Auth extends SafeEventEmitter<Web3AuthSfaEvents> implements IWeb3Auth {
    readonly coreOptions: Omit<Web3AuthOptions, "storage"> & {
        storage: IAsyncStorage | IStorage | ISecureStore;
        sessionKey?: string;
    };
    readonly connectedAdapterName: string;
    readonly SFA_ISSUER = "https://authjs.web3auth.io/jwks";
    status: ADAPTER_STATUS_TYPE;
    authInstance: Torus | null;
    state: SessionData;
    private nodeDetails;
    private torusPrivKey;
    private privKeyProvider;
    private accountAbstractionProvider;
    private sessionManager;
    private currentStorage;
    private readonly baseStorageKey;
    private readonly sessionNamespace;
    private plugins;
    constructor(options: Web3AuthOptions & {
        sessionKey?: string;
    });
    get connected(): boolean;
    get provider(): IProvider | null;
    set provider(_: IProvider | null);
    init(): Promise<void>;
    authenticateUser(): Promise<UserAuthInfo>;
    addChain(chainConfig: CustomChainConfig): Promise<void>;
    switchChain(params: {
        chainId: string;
    }): Promise<void>;
    getPostboxKey(loginParams: LoginParams): Promise<string>;
    /**
     * Use this function only with verifiers created on developer dashboard (https://dashboard.web3auth.io)
     * @param loginParams - Params used to login
     * @returns provider to connect
     */
    connect(loginParams: LoginParams): Promise<IProvider | null>;
    logout(): Promise<void>;
    getUserInfo(): Promise<AuthUserInfo>;
    addPlugin(plugin: IPlugin): IWeb3Auth;
    getPlugin(name: string): IPlugin | null;
    _finalizeLogin(params: IFinalizeLoginParams): Promise<void>;
    _getBasePrivKey(): string;
    private updateState;
    private getFinalPrivKey;
    private getSubKey;
    private getEd25519Key;
    private getTorusKey;
    private isSapphireNetwork;
    private getSessionSignatures;
    private subscribeToEvents;
    private connectToPlugins;
    private _getSignedMessage;
    private isRNOrNodeMode;
    private getStorage;
}
